#   Copyright (c) 2019 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import paddle.fluid as fluid


class FLJobBase(object):
    """
    FLJobBase is fl job base class, responsible for save and load
    a federated learning job
    """

    def __init__(self):
        pass

    def _save_str_list(self, items, output):
        with open(output, "w") as fout:
            for item in items:
                fout.write(item + "\n")

    def _load_str_list(self, input_file):
        res = []
        with open(input_file, "r") as fin:
            for line in fin:
                res.append(line.strip())
        return res

    def _save_strategy(self, strategy, output_file):
        import pickle
        pickle.dump(strategy, open(output_file, "wb"))

    def _save_endpoints(self, endpoints, output_file):
        with open(output_file, "w") as fout:
            for ep in endpoints:
                fout.write(str(ep) + "\n")

    def _load_endpoints(self, input_file):
        ep_list = []
        with open(input_file, "r") as fin:
            for line in fin:
                ep_list.append(line.strip())
        return ep_list

    def _save_program(self, program, output_file):
        with open(output_file, "wb") as fout:
            fout.write(program.desc.serialize_to_string())

    def _save_readable_program(self, program, output_file):
        with open(output_file, "w") as fout:
            fout.write(str(program))

    def _load_program(self, input_file):
        with open(input_file, "rb") as fin:
            program_desc_str = fin.read()
            return fluid.Program.parse_from_string(program_desc_str)
        return None


class FLCompileTimeJob(FLJobBase):
    """
    FLCompileTimeJob is a container for compile time job in federated learning.
    trainer startup programs, trainer main programs and other trainer programs
    are in FLCompileTimeJob. Also, server main programs and server startup programs
    are in this class. FLCompileTimeJob has server endpoints for debugging as well
    """

    def __init__(self):
        self._trainer_startup_programs = []
        self._trainer_recv_programs = []
        self._trainer_main_programs = []
        self._trainer_send_programs = []
        self._server_startup_programs = []
        self._server_main_programs = []
        self._server_endpoints = []

    def set_strategy(self, strategy):
        self._strategy = strategy

    def set_server_endpoints(self, ps_endpoints):
        self._server_endpoints = ps_endpoints

    def set_feed_names(self, names):
        self._feed_names = names

    def set_target_names(self, names):
        self._target_names = names

    def save(self, folder=None):
        server_num = len(self._server_startup_programs)
        trainer_num = len(self._trainer_startup_programs)
        send_prog_num = len(self._trainer_send_programs)
        for i in range(server_num):
            server_folder = "%s/server%d" % (folder, i)
            os.system("mkdir -p %s" % server_folder)
            server_startup = self._server_startup_programs[i]
            server_main = self._server_main_programs[i]
            self._save_program(server_startup,
                               "%s/server.startup.program" % server_folder)
            self._save_program(server_main,
                               "%s/server.main.program" % server_folder)
            self._save_readable_program(server_startup,
                                        "%s/server.startup.program.txt" %
                                        server_folder)
            self._save_readable_program(
                server_main, "%s/server.main.program.txt" % server_folder)
            self._save_str_list(self._feed_names,
                                "%s/feed_names" % server_folder)
            self._save_str_list(self._target_names,
                                "%s/target_names" % server_folder)
            self._save_endpoints(self._server_endpoints,
                                 "%s/endpoints" % server_folder)
            self._save_strategy(self._strategy,
                                "%s/strategy.pkl" % server_folder)

        for i in range(trainer_num):
            trainer_folder = "%s/trainer%d" % (folder, i)
            os.system("mkdir -p %s" % trainer_folder)
            trainer_startup = self._trainer_startup_programs[i]
            trainer_main = self._trainer_main_programs[i]
            self._save_program(trainer_startup,
                               "%s/trainer.startup.program" % trainer_folder)
            self._save_program(trainer_main,
                               "%s/trainer.main.program" % trainer_folder)
            self._save_readable_program(trainer_startup,
                                        "%s/trainer.startup.program.txt" %
                                        trainer_folder)
            self._save_readable_program(
                trainer_main, "%s/trainer.main.program.txt" % trainer_folder)
            self._save_str_list(self._feed_names,
                                "%s/feed_names" % trainer_folder)
            self._save_str_list(self._target_names,
                                "%s/target_names" % trainer_folder)
            self._save_endpoints(self._server_endpoints,
                                 "%s/endpoints" % trainer_folder)
            self._save_strategy(self._strategy,
                                "%s/strategy.pkl" % trainer_folder)

        for i in range(send_prog_num):
            trainer_folder = "%s/trainer%d" % (folder, i)
            trainer_send = self._trainer_send_programs[i]
            trainer_recv = self._trainer_recv_programs[i]
            self._save_program(trainer_send,
                               "%s/trainer.send.program" % trainer_folder)
            self._save_program(trainer_recv,
                               "%s/trainer.recv.program" % trainer_folder)
            self._save_readable_program(
                trainer_send, "%s/trainer.send.program.txt" % trainer_folder)
            self._save_readable_program(
                trainer_recv, "%s/trainer.recv.program.txt" % trainer_folder)


class FLRunTimeJob(FLJobBase):
    """
    FLRunTimeJob is a contrainer for run time job in federated leanring.
    A trainer or a server can load FLRunTimeJob. Only necessary programs
     can be loaded in FLRunTimeJob
    """

    def __init__(self):
        self._trainer_startup_program = None
        self._trainer_recv_program = None
        self._trainer_main_program = None
        self._trainer_send_program = None
        self._server_startup_program = None
        self._server_main_program = None
        self._feed_names = None
        self._target_names = None
        self._scheduler_ep = None

    def _load_strategy(self, input_file):
        import pickle
        return pickle.load(open(input_file, "rb"))

    def load_trainer_job(self, folder=None, trainer_id=0):
        """
        Load trainer job given training folder and trainer id
        Currently, a trainer_id is assigned to a trainer node, and
        corresponding FL Job will be sent to the trainer node.

        Args:
            folder(str): FL Job folder name
            trainer_id(int): trainer index for current job

        Return:
            None
        """
        folder_name = "%s/trainer%d" % (folder, trainer_id)
        startup_fn = "%s/trainer.startup.program" % folder_name
        self._trainer_startup_program = self._load_program(startup_fn)

        main_fn = "%s/trainer.main.program" % folder_name
        self._trainer_main_program = self._load_program(main_fn)

        try:
            send_fn = "%s/trainer.send.program" % folder_name
            self._trainer_send_program = self._load_program(send_fn)

            recv_fn = "%s/trainer.recv.program" % folder_name
            self._trainer_recv_program = self._load_program(recv_fn)
        except:
            pass

        endpoints_fn = "%s/endpoints" % folder_name
        self._endpoints = self._load_endpoints(endpoints_fn)

        strategy_fn = "%s/strategy.pkl" % folder_name
        self._strategy = self._load_strategy(strategy_fn)

        feed_names_fn = "%s/feed_names" % folder_name
        self._feed_names = self._load_str_list(feed_names_fn)

        target_names_fn = "%s/target_names" % folder_name
        self._target_names = self._load_str_list(target_names_fn)

    def load_server_job(self, folder=None, server_id=0):
        """
        Load server job given training folder and server_id
        Currently, a server_id is assigned to a server node, and
        corresponding FL Job will be sent to the server node.

        Args:
            folder(str): FL Job folder name
            server_id(int): server index for current job

        Return:
            None
        """
        folder_name = "%s/server%d" % (folder, server_id)
        startup_fn = "%s/server.startup.program" % folder_name
        self._server_startup_program = self._load_program(startup_fn)

        main_fn = "%s/server.main.program" % folder_name
        self._server_main_program = self._load_program(main_fn)

        endpoints_fn = "%s/endpoints" % folder_name
        self._endpoints = self._load_endpoints(endpoints_fn)

        import pickle
        strategy_fn = "%s/strategy.pkl" % folder_name
        self._strategy = self._load_strategy(strategy_fn)



# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import paddle.fluid as fluid
import numpy as np
import os


class Gru4rec_Reader:
    def __init__(self):
        pass

    def to_lodtensor(self, data, place):
        """ convert to LODtensor """
        seq_lens = [len(seq) for seq in data]
        cur_len = 0
        lod = [cur_len]
        for l in seq_lens:
            cur_len += l
            lod.append(cur_len)
        flattened_data = np.concatenate(data, axis=0).astype("int64")
        flattened_data = flattened_data.reshape([len(flattened_data), 1])
        res = fluid.LoDTensor()
        res.set(flattened_data, place)
        res.set_lod([lod])
        return res

    def lod_reader(self, reader, place):
        def feed_reader():
            for data in reader():
                lod_src_wordseq = self.to_lodtensor([dat[0] for dat in data],
                                                    place)
                lod_dst_wordseq = self.to_lodtensor([dat[1] for dat in data],
                                                    place)
                fe_data = {}
                fe_data["src_wordseq"] = lod_src_wordseq
                fe_data["dst_wordseq"] = lod_dst_wordseq
                yield fe_data

        return feed_reader

    def sort_batch(self, reader, batch_size, sort_group_size, drop_last=False):
        """
        Create a batched reader.
        """

        def batch_reader():
            r = reader()
            b = []
            for instance in r:
                b.append(instance)
                if len(b) == sort_group_size:
                    sortl = sorted(b, key=lambda x: len(x[0]), reverse=True)
                    b = []
                    c = []
                    for sort_i in sortl:
                        c.append(sort_i)
                        if (len(c) == batch_size):
                            yield c
                            c = []

            if drop_last == False and len(b) != 0:
                sortl = sorted(b, key=lambda x: len(x[0]), reverse=True)
                c = []
                for sort_i in sortl:
                    c.append(sort_i)
            if (len(c) == batch_size):
                yield c
                c = []

        # Batch size check
        batch_size = int(batch_size)
        if batch_size <= 0:
            raise ValueError(
                "batch_size should be a positive integeral value, "
                "but got batch_size={}".format(batch_size))
        return batch_reader

    def reader_creator(self, file_dir):
        def reader():
            files = os.listdir(file_dir)
            for fi in files:
                with open(file_dir + '/' + fi, "r") as f:
                    for l in f:
                        l = l.strip().split()
                        l = [w for w in l]
                        src_seq = l[:len(l) - 1]
                        trg_seq = l[1:]
                        yield src_seq, trg_seq

        return reader

    def reader(self, file_dir, place, batch_size=5):
        """ prepare the English Pann Treebank (PTB) data """
        print("start constuct word dict")
        reader = self.sort_batch(
            self.reader_creator(file_dir), batch_size, batch_size * 20)
        return self.lod_reader(reader, place)



# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import requests
import os
import json
import tarfile
import random


def download(url, tar_path):
    r = requests.get(url)
    with open(tar_path, 'wb') as f:
        f.write(r.content)


def extract(tar_path, target_path):
    tar = tarfile.open(tar_path, "r:gz")
    file_names = tar.getnames()
    for file_name in file_names:
        tar.extract(file_name, target_path)

    tar.close()


def train(trainer_id, inner_step, batch_size, count_by_step):
    target_path = "trainer%d_data" % trainer_id
    data_path = target_path + "/femnist_data"
    tar_path = data_path + ".tar.gz"
    if not os.path.exists(target_path):
        os.system("mkdir trainer%d_data" % trainer_id)
    if not os.path.exists(data_path):
        print("Preparing data...")
        if not os.path.exists(tar_path):
            download("https://paddlefl.bj.bcebos.com/leaf/femnist_data.tar.gz",
                     tar_path)
            extract(tar_path, target_path)

    def train_data():
        train_file = open(
            "./trainer%d_data/femnist_data/train/all_data_%d_niid_0_keep_0_train_9.json"
            % (trainer_id, trainer_id), 'r')
        json_train = json.load(train_file)
        users = json_train["users"]
        rand = random.randrange(
            0, len(users))  # random choose a user from each trainer
        cur_user = users[rand]
        print('training using ' + cur_user)
        train_images = json_train["user_data"][cur_user]['x']
        train_labels = json_train["user_data"][cur_user]['y']
        if count_by_step:
            for i in range(inner_step * batch_size):
                yield train_images[i % (len(train_images))], train_labels[i % (
                    len(train_images))]
        else:
            for i in range(len(train_images)):
                yield train_images[i], train_labels[i]

        train_file.close()

    return train_data


def test(trainer_id, inner_step, batch_size, count_by_step):
    target_path = "trainer%d_data" % trainer_id
    data_path = target_path + "/femnist_data"
    tar_path = data_path + ".tar.gz"
    if not os.path.exists(target_path):
        os.system("mkdir trainer%d_data" % trainer_id)
    if not os.path.exists(data_path):
        print("Preparing data...")
        if not os.path.exists(tar_path):
            download("https://paddlefl.bj.bcebos.com/leaf/femnist_data.tar.gz",
                     tar_path)
        extract(tar_path, target_path)

    def test_data():
        test_file = open(
            "./trainer%d_data/femnist_data/test/all_data_%d_niid_0_keep_0_test_9.json"
            % (trainer_id, trainer_id), 'r')
        json_test = json.load(test_file)
        users = json_test["users"]
        for user in users:
            test_images = json_test['user_data'][user]['x']
            test_labels = json_test['user_data'][user]['y']
            for i in range(len(test_images)):
                yield test_images[i], test_labels[i]

        test_file.close()

    return test_data



# Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import six

from paddle.fluid import core
import paddle


def delete_ops(block, ops):
    for op in ops:
        try:
            idx = list(block.ops).index(op)
            block._remove_op(idx)
        except Exception as e:
            print(e)


def find_op_by_input_arg(block, arg_name):
    for index, op in enumerate(block.ops):
        if arg_name in op.input_arg_names:
            return index
    return -1


def find_op_by_output_arg(block, arg_name, reverse=False):
    if reverse:
        pos = len(block.ops) - 1
        while pos >= 0:
            op = block.ops[pos]
            if arg_name in op.output_arg_names:
                return pos
            pos -= 1
    else:
        for index, op in enumerate(block.ops):
            if arg_name in op.output_arg_names:
                return index
    return -1


def get_indent_space(indent, space_num=4):
    ret = ""
    for i in range(0, indent * space_num):
        ret += " "

    return ret


def variable_to_code(var):
    """
    Get readable codes of fluid variable.

    Args:
        var: A fluid operator.

    Returns:
        string: The formatted string.
    """
    if var.type == core.VarDesc.VarType.SELECTED_ROWS or var.type == core.VarDesc.VarType.LOD_TENSOR:
        var_str = "{name} : fluid.{type}.shape{shape}.astype({dtype})".\
            format(i="{", e="}", name=var.name, type=var.type, shape=var.shape, dtype=var.dtype)
    else:
        var_str = "{name} : fluid.{type})".\
            format(i="{", e="}", name=var.name, type=var.type)

    if type(var) == paddle.fluid.framework.Parameter:
        if var.trainable:
            var_str = "trainable parameter " + var_str
        else:
            var_str = "parameter " + var_str
    else:
        var_str = "var " + var_str

    if var.persistable:
        var_str = "persist " + var_str

    return var_str


def op_to_code(op, skip_op_callstack=True):
    """
    Get readable codes of fluid operator.

    Args:
        op: A fluid operator.

    Returns:
        string: The foramtted string.
    """

    outputs_str = "{"
    for i in range(0, len(op.output_names)):
        outputs_str += "{name}=".format(name=op.output_names[i])
        o = op.output(op.output_names[i])
        outputs_str += "{value}".format(value=o)
        if i != len(op.output_names) - 1:
            outputs_str += ", "
    outputs_str += "}"

    inputs_str = "{"
    for i in range(0, len(op.input_names)):
        inputs_str += "{name}=".format(name=op.input_names[i])
        o = op.input(op.input_names[i])
        inputs_str += "{value}".format(value=o)

        if i != len(op.input_names) - 1:
            inputs_str += ", "
    inputs_str += "}"

    attr_names = sorted(op.attr_names)
    attrs_str = ""
    for i in range(0, len(attr_names)):
        name = attr_names[i]
        if skip_op_callstack and name == "op_callstack":
            continue

        attr_type = op.desc.attr_type(name)
        if attr_type == core.AttrType.BLOCK:
            a = "{name} = block[{value}]".format(
                name=name, type=attr_type, value=op._block_attr_id(name))
            attrs_str += a
            if i != len(attr_names) - 1:
                attrs_str += ", "
            continue

        if attr_type == core.AttrType.BLOCKS:
            a = "{name} = blocks{value}".format(
                name=name, type=attr_type, value=op._blocks_attr_ids(name))
            attrs_str += a
            if i != len(attr_names) - 1:
                attrs_str += ", "
            continue

        a = "{name} = {value}".format(
            name=name, type=attr_type, value=op.desc.attr(name))
        attrs_str += a
        if i != len(attr_names) - 1:
            attrs_str += ", "

    if outputs_str != "{}":
        op_str = "{outputs} = {op_type}(inputs={inputs}, {attrs})".\
            format(outputs = outputs_str, op_type=op.type, inputs=inputs_str, attrs=attrs_str)
    else:
        op_str = "{op_type}(inputs={inputs}, {attrs})".\
            format(op_type=op.type, inputs=inputs_str, attrs=attrs_str)
    return op_str


def block_to_code(block, block_idx, fout=None, skip_op_callstack=False):
    indent = 0

    print(
        "{0}{1} // block {2}".format(
            get_indent_space(indent), '{', block_idx),
        file=fout)

    indent += 1
    # sort all vars
    all_vars = sorted(six.iteritems(block.vars), key=lambda x: x[0])
    for var in all_vars:
        print(
            "{}{}".format(get_indent_space(indent), variable_to_code(var[1])),
            file=fout)

    if len(all_vars) > 0:
        print("", file=fout)

    for op in block.ops:
        print(
            "{}{}".format(
                get_indent_space(indent), op_to_code(op, skip_op_callstack)),
            file=fout)
    indent -= 1

    print("{0}{1}".format(get_indent_space(indent), '}'), file=fout)


def program_to_code(prog, fout=None, skip_op_callstack=True):
    """
    Print readable codes of fluid program.

    Args:
        prog : A fluid program.

    An example result like bellow:
    https://github.com/PaddlePaddle/Paddle/pull/12673
    """
    block_idx = 0
    for block in prog.blocks:
        block_to_code(block, block_idx, fout, skip_op_callstack)
        block_idx += 1



#   Copyright (c) 2019 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import paddle.fluid as fluid
from .fl_job import FLCompileTimeJob


class JobGenerator(object):
    """
    A JobGenerator is responsible for generating distributed federated
    learning configs. Before federated learning job starts, organizations
    need to define a deep learning model together to do horizontal federated
    learning.
    """

    def __init__(self):
        # worker num for federated learning
        self._worker_num = 0
        # startup program
        self._startup_prog = None
        # inner optimizer
        self._optimizer = \
            fluid.optimizer.SGD(learning_rate=0.001)
        self._feed_names = []
        self._target_names = []

    def set_optimizer(self, optimizer):
        """
        Set optimizer of current job
        """
        self._optimizer = optimizer

    def set_losses(self, losses):
        """
        Set losses of current job
        losses can be a list of loss so that we can do
        optimization on multiple losses
        """
        self._losses = losses

    def set_startup_program(self, startup=None):
        """
        set startup program for user defined program
        """
        if startup == None:
            startup = fluid.default_startup_program()
        self._startup_prog = startup

    def set_infer_feed_and_target_names(self, feed_names, target_names):
        if not isinstance(feed_names, list) or not isinstance(target_names,
                                                              list):
            raise ValueError(
                "input should be list in set_infer_feed_and_target_names")
        '''
        print(feed_names)
        print(target_names)
        for item in feed_names:
            if type(item) != str:
                raise ValueError("item in feed_names should be string")
        for item in target_names:
            if type(item) != str:
                raise ValueError("item in target_names should be string")
        '''
        self._feed_names = feed_names
        self._target_names = target_names

    def generate_fl_job(self,
                        fl_strategy,
                        server_endpoints=[],
                        worker_num=1,
                        output=None):
        """
        Generate Federated Learning Job, based on user defined configs

        Args:
            fl_strategy(FLStrategyBase): federated learning strategy defined by current federated users
            server_endpoints(List(str)): endpoints for federated server nodes
            worker_num(int): number of training nodes
            output(str): output directory of generated fl job

        Returns:
            None

        Examples:
            import paddle.fluid as fluid
            import paddle_fl as fl
            from paddle_fl.core.master.job_generator import JobGenerator
            from paddle_fl.core.strategy.fl_strategy_base import FLStrategyFactory

            input_x = fluid.layers.data(name="input_x", shape=[10], dtype="float32")
            label = fluid.layers.data(name="label", shape[1], dtype="int64")
            fc0 = fluid.layers.fc(input=input_x, size=2, act='sigmoid')
            cost = fluid.layers.cross_entropy(input=fc0, label=label)
            loss = fluid.layers.reduce_mean(cost)

            job_generator = JobGenerator()
            optimizer = fluid.optimizer.SGD(learning_rate=0.1)
            job_generator.set_optimizer(optimizer)
            job_generator.set_losses([loss])
            server_endpoints = [127.0.0.1:8181]
            worker_num = 10
            build_strategy = FLStrategyFactor()
            build_strategy.fed_avg = True
            strategy = build_strategy.create_fl_strategy()
            job_output_dir = "fl_job_config"
            job_generator.generate_fl_job(strategy,
                                          server_endpoints=server_endpoints,
                                          worker_num=1,
                                          output=output)

        """
        local_job = FLCompileTimeJob()
        assert len(self._losses) > 0
        assert self._startup_prog != None
        assert fl_strategy != None
        assert output != None
        fl_strategy.minimize(self._optimizer, self._losses)

        # strategy can generate startup and main program
        # of a single worker and servers
        for trainer_id in range(worker_num):
            startup_program = self._startup_prog.clone()
            main_program = self._losses[0].block.program.clone()
            fl_strategy._build_trainer_program_for_job(
                trainer_id,
                program=main_program,
                ps_endpoints=server_endpoints,
                trainers=worker_num,
                sync_mode=True,
                startup_program=startup_program,
                job=local_job)

        startup_program = self._startup_prog.clone()
        main_program = self._losses[0].block.program.clone()
        fl_strategy._build_server_programs_for_job(
            program=main_program,
            ps_endpoints=server_endpoints,
            trainers=worker_num,
            sync_mode=True,
            startup_program=startup_program,
            job=local_job)

        local_job.set_feed_names(self._feed_names)
        local_job.set_target_names(self._target_names)
        local_job.set_strategy(fl_strategy)
        local_job.save(output)

    def generate_fl_job_for_k8s(self,
                                fl_strategy,
                                server_pod_endpoints=[],
                                server_service_endpoints=[],
                                worker_num=1,
                                output=None):

        local_job = FLCompileTimeJob()
        assert len(self._losses) > 0
        assert self._startup_prog != None
        assert fl_strategy != None
        assert output != None
        fl_strategy.minimize(self._optimizer, self._losses)

        # strategy can generate startup and main program
        # of a single worker and servers
        for trainer_id in range(worker_num):
            startup_program = self._startup_prog.clone()
            main_program = self._losses[0].block.program.clone()
            fl_strategy._build_trainer_program_for_job(
                trainer_id,
                program=main_program,
                ps_endpoints=server_service_endpoints,
                trainers=worker_num,
                sync_mode=True,
                startup_program=startup_program,
                job=local_job)

        startup_program = self._startup_prog.clone()
        main_program = self._losses[0].block.program.clone()
        fl_strategy._build_server_programs_for_job(
            program=main_program,
            ps_endpoints=server_pod_endpoints,
            trainers=worker_num,
            sync_mode=True,
            startup_program=startup_program,
            job=local_job)

        local_job.set_feed_names(self._feed_names)
        local_job.set_target_names(self._target_names)
        local_job.set_strategy(fl_strategy)
        local_job.save(output)



# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import os


class CloudClient(object):
    def __init__(self):
        pass

    def generate_submit_sh(self, job_dir):
        with open() as fout:
            pass

    def generate_job_sh(self, job_dir):
        with open() as fout:
            pass

    def submit(self, **kwargs):
        pass


class HPCClient(object):
    def __init__(self):
        self.conf_dict = {}

    def print_args(self):
        print("task_name: {}".format(self.task_name))
        print("hdfs_path: {}".format(self.hdfs_path))
        print("ugi: {}".format(self.ugi))
        print("hdfs_output: {}".format(self.hdfs_output))
        print("worker_nodes: {}".format(self.worker_nodes))
        print("server_nodes: {}".format(self.server_nodes))
        print("hadoop_home: {}".format(self.hadoop_home))
        print("hpc_home: {}".format(self.hpc_home))
        print("train_cmd: {}".format(self.train_cmd))
        print("package_path: {}".format(self.package_path))
        print("priority: {}".format(self.priority))
        print("queue: {}".format(self.queue))
        print("server: {}".format(self.server))
        print("mpi_node_mem: {}".format(self.mpi_node_mem))
        print("pcpu: {}".format(self.pcpu))
        print("python_tar: {}".format(self.python_tar))
        print("wheel: {}".format(self.wheel))

    def check_args(self):
        assert self.task_name != ""
        assert self.hdfs_path != ""
        assert self.ugi != ""
        assert self.hdfs_output != ""
        assert self.worker_nodes != ""
        assert self.server_nodes != ""
        assert self.hadoop_home != ""
        assert self.hpc_home != ""
        assert self.train_cmd != ""
        assert self.package_path != ""
        assert self.priority != ""
        assert self.queue != ""
        assert self.server != ""
        assert self.mpi_node_mem != ""
        assert self.pcpu != ""
        assert self.python_tar != ""
        assert self.wheel != ""

    def generate_qsub_conf(self, job_dir):
        with open("{}/qsub.conf".format(job_dir), "w") as fout:
            fout.write("SERVER={}\n".format(self.server))
            fout.write("QUEUE={}\n".format(self.queue))
            fout.write("PRIORITY={}\n".format(self.priority))
            fout.write("USE_FLAGS_ADVRES=yes\n")

    def generate_submit_sh(self, job_dir):
        with open("{}/submit.sh".format(job_dir), "w") as fout:
            fout.write("#!/bin/bash\n")
            fout.write("unset http_proxy\n")
            fout.write("unset https_proxy\n")
            fout.write("export HADOOP_HOME={}\n".format(self.hadoop_home))
            fout.write("$HADOOP_HOME/bin/hadoop fs -Dhadoop.job.ugi={}"
                       " -Dfs.default.name={} -rmr {}\n".format(
                           self.ugi, self.hdfs_path, self.hdfs_output))
            fout.write("MPI_NODE_MEM={}\n".format(self.mpi_node_mem))
            fout.write("{}/bin/qsub_f -N {} --conf qsub.conf "
                       "--hdfs {} --ugi {} --hout {} --files ./package "
                       "-l nodes={},walltime=1000:00:00,pmem-hard={},"
                       "pcpu-soft={},pnetin-soft=1000,"
                       "pnetout-soft=1000 job.sh\n".format(
                           self.hpc_home, self.task_name, self.hdfs_path,
                           self.ugi, self.hdfs_output,
                           int(self.worker_nodes) + int(self.server_nodes),
                           self.mpi_node_mem, self.pcpu))

    def generate_job_sh(self, job_dir):
        with open("{}/job.sh".format(job_dir), "w") as fout:
            fout.write("#!/bin/bash\n")
            fout.write("WORKDIR=`pwd`\n")
            fout.write("mpirun -npernode 1 mv package/* ./\n")
            fout.write("echo 'current dir: '$WORKDIR\n")
            fout.write(
                "mpirun -npernode 1 tar -zxvf python.tar.gz > /dev/null\n")
            fout.write(
                "export LIBRARY_PATH=$WORKDIR/python/lib:$LIBRARY_PATH\n")
            fout.write("mpirun -npernode 1 python/bin/python -m pip install "
                       "{} --index-url=http://pip.baidu.com/pypi/simple "
                       "--trusted-host pip.baidu.com > /dev/null\n".format(
                           self.wheel))
            fout.write("export PATH=python/bin:$PATH\n")
            if self.monitor_cmd != "":
                fout.write(
                    "mpirun -npernode 1 -timestamp-output -tag-output -machinefile "
                    "${{PBS_NODEFILE}} python/bin/{} > monitor.log 2> monitor.elog &\n".
                    format(self.monitor_cmd))
            fout.write(
                "mpirun -npernode 1 -timestamp-output -tag-output -machinefile ${PBS_NODEFILE} python/bin/python train_program.py\n"
            )
            fout.write("if [[ $? -ne 0 ]]; then\n")
            fout.write("    echo 'Failed to run mpi!' 1>&2\n")
            fout.write("    exit 1\n")
            fout.write("fi\n")

    def submit(self, **kwargs):
        # task_name, output_path
        self.task_name = kwargs.get("task_name", "test_submit_job")
        self.hdfs_path = kwargs.get("hdfs_path", "")
        self.ugi = kwargs.get("ugi", "")
        self.hdfs_output = kwargs.get("hdfs_output", "")
        self.worker_nodes = str(kwargs.get("worker_nodes", 2))
        self.server_nodes = str(kwargs.get("server_nodes", 2))
        self.hadoop_home = kwargs.get("hadoop_home", "")
        self.hpc_home = kwargs.get("hpc_home", "")
        self.train_cmd = kwargs.get("train_cmd", "")
        self.monitor_cmd = kwargs.get("monitor_cmd", "")
        self.package_path = kwargs.get("package_path", "")
        self.priority = kwargs.get("priority", "")
        self.queue = kwargs.get("queue", "")
        self.server = kwargs.get("server", "")
        self.mpi_node_mem = str(kwargs.get("mpi_node_mem", 11000))
        self.pcpu = str(kwargs.get("pcpu", 180))
        self.python_tar = kwargs.get("python_tar", "")
        self.wheel = kwargs.get("wheel", "")

        self.print_args()
        self.check_args()
        jobdir = "{}_jobdir".format(self.task_name)
        os.system("mkdir -p {}_jobdir".format(self.task_name))
        os.system("rm -rf {}/package".format(jobdir))
        os.system("cp -r {} {}/package".format(self.package_path, jobdir))
        os.system("cp {} {}/package/".format(self.python_tar, jobdir))
        os.system("cp {} {}/package/".format(self.wheel, jobdir))
        # make submit dir
        self.generate_submit_sh(jobdir)
        # generate submit.sh
        self.generate_job_sh(jobdir)
        # generate job.sh
        self.generate_qsub_conf(jobdir)
        # run submit
        os.system("cd {};sh submit.sh > submit.log 2> submit.elog &".format(
            jobdir))



# Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function


class UnionFind(object):
    """ Union-find data structure.

    Union-find is a data structure that keeps track of a set of elements partitioned
    into a number of disjoint (non-overlapping) subsets.

    Reference:
    https://en.wikipedia.org/wiki/Disjoint-set_data_structure

    Args:
      elements(list): The initialize element list.
    """

    def __init__(self, elementes=None):
        self._parents = []  # index -> parent index
        self._index = {}  # element -> index
        self._curr_idx = 0
        if not elementes:
            elementes = []
        for ele in elementes:
            self._parents.append(self._curr_idx)
            self._index.update({ele: self._curr_idx})
            self._curr_idx += 1

    def find(self, x):
        # Find the root index of given element x,
        # execute the path compress while findind the root index
        if not x in self._index:
            return -1
        idx = self._index[x]
        while idx != self._parents[idx]:
            t = self._parents[idx]
            self._parents[idx] = self._parents[t]
            idx = t
        return idx

    def union(self, x, y):
        # Union two given element
        x_root = self.find(x)
        y_root = self.find(y)

        if x_root == y_root:
            return
        self._parents[x_root] = y_root

    def is_connected(self, x, y):
        # If two given elements have the same root index,
        # then they are connected.
        return self.find(x) == self.find(y)



# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import zmq
import time
import random


def recv_and_parse_kv(socket):
    message = socket.recv()
    group = message.decode().split("\t")
    if group[0] == "alive":
        return group[0], "0"
    else:
        return group[0], group[1]


WORKER_EP = "WORKER_EP"
SERVER_EP = "SERVER_EP"


class FLServerAgent(object):
    def __init__(self, scheduler_ep, current_ep):
        self.scheduler_ep = scheduler_ep
        self.context = zmq.Context()
        self.socket = self.context.socket(zmq.REQ)
        self.socket.connect("tcp://{}".format(scheduler_ep))
        self.current_ep = current_ep

    def connect_scheduler(self):
        while True:
            self.socket.send_string("SERVER_EP\t{}".format(self.current_ep))
            message = self.socket.recv()
            group = message.decode().split("\t")
            if group[0] == 'INIT':
                break


class FLWorkerAgent(object):
    def __init__(self, scheduler_ep, current_ep):
        self.scheduler_ep = scheduler_ep
        self.context = zmq.Context()
        self.socket = self.context.socket(zmq.REQ)
        self.socket.connect("tcp://{}".format(scheduler_ep))
        self.current_ep = current_ep

    def connect_scheduler(self):
        while True:
            self.socket.send_string("WORKER_EP\t{}".format(self.current_ep))
            message = self.socket.recv()
            group = message.decode().split("\t")
            if group[0] == 'INIT':
                break

    def finish_training(self):
        self.socket.send_string("FINISH\t{}".format(self.current_ep))
        key, value = recv_and_parse_kv(self.socket)
        if key == "WAIT":
            time.sleep(3)
            return True
        return False

    def can_join_training(self):
        self.socket.send_string("JOIN\t{}".format(self.current_ep))
        key, value = recv_and_parse_kv(self.socket)

        if key == "ACCEPT":
            return True
        elif key == "REJECT":
            return False
        return False


class FLScheduler(object):
    def __init__(self, worker_num, server_num, port=9091, socket=None):
        self.context = zmq.Context()
        if socket == None:
            self.socket = self.context.socket(zmq.REP)
            self.socket.bind("tcp://*:{}".format(port))
        else:
            self.socket = socket
        self.worker_num = worker_num
        self.server_num = server_num
        self.sample_worker_num = 0
        self.fl_workers = []
        self.fl_servers = []

    def set_sample_worker_num(self, sample_worker_num=0):
        if sample_worker_num == 0:
            self.sample_worker_num = int(self.worker_num * 0.1)
        else:
            self.sample_worker_num = sample_worker_num

    def init_env(self):
        ready = False
        while not ready:
            key, value = recv_and_parse_kv(self.socket)
            if key == WORKER_EP:
                self.fl_workers.append(value)
                self.socket.send_string("INIT\t{}".format(value))
            elif key == SERVER_EP:
                self.fl_servers.append(value)
                self.socket.send_string("INIT\t{}".format(value))
            else:
                time.sleep(3)
                self.socket.send_string("REJECT\t0")
            if len(self.fl_workers) == self.worker_num and \
               len(self.fl_servers) == self.server_num:
                ready = True

    def start_fl_training(self):
        # loop until training is done
        while True:
            random.shuffle(self.fl_workers)
            worker_dict = {}
            for worker in self.fl_workers[:self.sample_worker_num]:
                worker_dict[worker] = 0

            ready_workers = []
            all_ready_to_train = False
            while not all_ready_to_train:
                key, value = recv_and_parse_kv(self.socket)
                if key == "JOIN":
                    if value in worker_dict:
                        if worker_dict[value] == 0:
                            ready_workers.append(value)
                            worker_dict[value] = 1
                            self.socket.send_string("ACCEPT\t0")
                            continue
                    else:
                        if value not in ready_workers:
                            ready_workers.append(value)
                self.socket.send_string("REJECT\t0")
                if len(ready_workers) == len(self.fl_workers):
                    all_ready_to_train = True

            all_finish_training = False
            finish_training_dict = {}
            while not all_finish_training:
                key, value = recv_and_parse_kv(self.socket)
                if key == "FINISH":
                    finish_training_dict[value] = 1
                    self.socket.send_string("WAIT\t0")
                else:
                    self.socket.send_string("REJECT\t0")
                if len(finish_training_dict) == len(worker_dict):
                    all_finish_training = True
            time.sleep(5)



#   Copyright (c) 2019 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import paddle.fluid as fluid
from paddle_fl.core.scheduler.agent_master import FLServerAgent


class FLServer(object):
    def __init__(self):
        self._startup_program = None
        self._main_program = None
        self._scheduler_ep = None
        self._current_ep = None

    def set_server_job(self, job):
        # need to parse startup and main program in job
        # need to parse current endpoint
        # need to parse master endpoint
        self._startup_program = job._server_startup_program
        self._main_program = job._server_main_program
        self._scheduler_ep = job._scheduler_ep
        self._current_ep = None

    def start(self):
        self.agent = FLServerAgent(self._scheduler_ep, self._current_ep)
        self.agent.connect_scheduler()
        exe = fluid.Executor(fluid.CPUPlace())
        exe.run(self._startup_program)
        exe.run(self._main_program)



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function


class PSDispatcher(object):
    """
    PSDispatcher is the base class for dispatching vars
    into different pserver instance.
    You need to implement the `dispatch` inferface.
    """

    def __init__(self, pserver_endpoints):
        self._eps = pserver_endpoints
        self._step = 0

    @property
    def eps(self):
        return self._eps

    def reset(self):
        self._step = 0

    def dispatch(self, varlist):
        """
        Args:
            varlist(list): a list of Variables
        Returns:
            a map of pserver endpoint -> varname
        """
        AssertionError("Interface has not been implemented.")


class HashName(PSDispatcher):
    """
    Hash variable names to several endpoints using python
    "hash()" function.

    Args:
        pserver_endpoints (list): list of endpoint(ip:port).

    Examples:
        .. code-block:: python

        pserver_endpoints = ["127.0.0.1:6007", "127.0.0.1:6008"]
        vars = ["var1","var2","var3","var4","var5"]

        rr = RoundRobin(pserver_endpoints)
        rr.dispatch(vars)

    """

    def __init__(self, pserver_endpoints):
        super(self.__class__, self).__init__(pserver_endpoints)

    def _hash_block(self, block_str, total):
        return hash(block_str) % total

    def dispatch(self, varlist):
        eplist = []
        for var in varlist:
            server_id = self._hash_block(var.name(), len(self._eps))
            server_for_param = self._eps[server_id]
            eplist.append(server_for_param)
        return eplist


class RoundRobin(PSDispatcher):
    """
    Distribute variables to serveral endpoints using
    RondRobin<https://en.wikipedia.org/wiki/Round-robin_scheduling> method.

    Args:
        pserver_endpoints (list): list of endpoint(ip:port).

    Examples:
        .. code-block:: python

        pserver_endpoints = ["127.0.0.1:6007", "127.0.0.1:6008"]
        vars = ["var1","var2","var3","var4","var5"]

        rr = RoundRobin(pserver_endpoints)
        rr.dispatch(vars)

    """

    def __init__(self, pserver_endpoints):
        super(self.__class__, self).__init__(pserver_endpoints)

    def dispatch(self, varlist):
        eplist = []
        for var in varlist:
            server_for_param = self._eps[self._step]
            eplist.append(server_for_param)
            self._step += 1
            if self._step >= len(self._eps):
                self._step = 0
        return eplist



# coding=utf-8

#
# (c) Chris von Csefalvay, 2015.
"""
exceptions is responsible for exception handling etc.
"""


class MalformedPublicKey(BaseException):
    """
    The public key is malformed as it does not meet the Legendre symbol criterion. The key might have been tampered with or might have been damaged in transit.
    """

    def __str__(self):
        return "Public key malformed: fails Legendre symbol verification."


class RNGError(BaseException):
    """
    Thrown when RNG could not be obtained.
    """

    def __str__(self):
        return "RNG could not be obtained. This module currently only works with Python 3."



#   Copyright (c) 2019 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import paddle.fluid as fluid
import logging
from paddle_fl.core.scheduler.agent_master import FLWorkerAgent
import numpy
import hmac
import hashlib
from .diffiehellman.diffiehellman import DiffieHellman


class FLTrainerFactory(object):
    def __init__(self):
        pass

    def create_fl_trainer(self, job):
        strategy = job._strategy
        trainer = None
        if strategy._fed_avg == True:
            trainer = FedAvgTrainer()
            trainer.set_trainer_job(job)
        elif strategy._dpsgd == True:
            trainer = FLTrainer()
            trainer.set_trainer_job(job)
        elif strategy._sec_agg == True:
            trainer = SecAggTrainer()
            trainer.set_trainer_job(job)
        trainer.set_trainer_job(job)
        return trainer


class FLTrainer(object):
    def __init__(self):
        self._logger = logging.getLogger("FLTrainer")
        pass

    def set_trainer_job(self, job):
        self._startup_program = \
            job._trainer_startup_program
        self._main_program = \
            job._trainer_main_program
        self._step = job._strategy._inner_step
        self._feed_names = job._feed_names
        self._target_names = job._target_names
        self._scheduler_ep = job._scheduler_ep
        self._current_ep = None
        self.cur_step = 0

    def start(self):
        #current_ep = "to be added"
        self.agent = FLWorkerAgent(self._scheduler_ep, self._current_ep)
        self.agent.connect_scheduler()
        self.exe = fluid.Executor(fluid.CPUPlace())
        self.exe.run(self._startup_program)

    def run(self, feed, fetch):
        self._logger.debug("begin to run")
        self.exe.run(self._main_program, feed=feed, fetch_list=fetch)
        self._logger.debug("end to run current batch")
        self.cur_step += 1

    def save_inference_program(self, output_folder):
        target_vars = []
        infer_program = self._main_program.clone(for_test=True)
        for name in self._target_names:
            tmp_var = self._main_program.block(0)._find_var_recursive(name)
            target_vars.append(tmp_var)
        fluid.io.save_inference_model(
            output_folder,
            self._feed_names,
            target_vars,
            self.exe,
            main_program=infer_program)

    def stop(self):
        # ask for termination with master endpoint
        # currently not open sourced, will release the code later
        # TODO(guru4elephant): add connection with master
        if self.cur_step != 0:
            while not self.agent.finish_training():
                self._logger.debug("Wait others finish")
                continue
        while not self.agent.can_join_training():
            self._logger.debug("Wait permit")
            continue
        self._logger.debug("Ready to train")
        return False


class FedAvgTrainer(FLTrainer):
    def __init__(self):
        super(FedAvgTrainer, self).__init__()
        pass

    def start(self):
        #current_ep = "to be added"
        self.agent = FLWorkerAgent(self._scheduler_ep, self._current_ep)
        self.agent.connect_scheduler()
        self.exe = fluid.Executor(fluid.CPUPlace())
        self.exe.run(self._startup_program)

    def set_trainer_job(self, job):
        super(FedAvgTrainer, self).set_trainer_job(job)
        self._send_program = job._trainer_send_program
        self._recv_program = job._trainer_recv_program

    def reset(self):
        self.cur_step = 0

    def run_with_epoch(self, reader, feeder, fetch, num_epoch):
        self._logger.debug("begin to run recv program")
        self.exe.run(self._recv_program)
        epoch = 0
        for i in range(num_epoch):
            for data in reader():
                self.exe.run(self._main_program,
                             feed=feeder.feed(data),
                             fetch_list=fetch)
            self.cur_step += 1
            epoch += 1
        self._logger.debug("begin to run send program")
        self.exe.run(self._send_program)

    def run(self, feed, fetch):
        self._logger.debug(
            "begin to run FedAvgTrainer, cur_step=%d, inner_step=%d" %
            (self.cur_step, self._step))
        if self.cur_step % self._step == 0:
            self._logger.debug("begin to run recv program")
            self.exe.run(self._recv_program)
        self._logger.debug("begin to run current step")
        loss = self.exe.run(self._main_program, feed=feed, fetch_list=fetch)
        if self.cur_step % self._step == 0:
            self._logger.debug("begin to run send program")
            self.exe.run(self._send_program)
        self.cur_step += 1
        return loss


class SecAggTrainer(FLTrainer):
    def __init__(self):
        super(SecAggTrainer, self).__init__()
        pass

    @property
    def trainer_id(self):
        return self._trainer_id

    @trainer_id.setter
    def trainer_id(self, s):
        self._trainer_id = s

    @property
    def trainer_num(self):
        return self._trainer_num

    @trainer_num.setter
    def trainer_num(self, s):
        self._trainer_num = s

    @property
    def key_dir(self):
        return self._key_dir

    @key_dir.setter
    def key_dir(self, s):
        self._key_dir = s

    @property
    def step_id(self):
        return self._step_id

    @step_id.setter
    def step_id(self, s):
        self._step_id = s

    def start(self):
        self.agent = FLWorkerAgent(self._scheduler_ep, self._current_ep)
        self.agent.connect_scheduler()
        self.exe = fluid.Executor(fluid.CPUPlace())
        self.exe.run(self._startup_program)
        self.cur_step = 0

    def set_trainer_job(self, job):
        super(SecAggTrainer, self).set_trainer_job(job)
        self._send_program = job._trainer_send_program
        self._recv_program = job._trainer_recv_program
        self_step = job._strategy._inner_step
        self._param_name_list = job._strategy._param_name_list

    def reset(self):
        self.cur_step = 0

    def run(self, feed, fetch):
        self._logger.debug(
            "begin to run SecAggTrainer, cur_step=%d, inner_step=%d" %
            (self.cur_step, self._step))
        if self.cur_step % self._step == 0:
            self._logger.debug("begin to run recv program")
            self.exe.run(self._recv_program)
        scope = fluid.global_scope()
        self._logger.debug("begin to run current step")
        loss = self.exe.run(self._main_program, feed=feed, fetch_list=fetch)
        if self.cur_step % self._step == 0:
            self._logger.debug("begin to run send program")
            noise = 0.0
            scale = pow(10.0, 5)
            digestmod = hashlib.sha256
            # 1. load priv key and other's pub key
            dh = DiffieHellman(group=15, key_length=256)
            dh.load_private_key(self._key_dir + str(self._trainer_id) +
                                "_priv_key.txt")
            key = str(self._step_id).encode("utf-8")
            for i in range(self._trainer_num):
                if i != self._trainer_id:
                    f = open(self._key_dir + str(i) + "_pub_key.txt", "r")
                    public_key = int(f.read())
                    dh.generate_shared_secret(public_key, echo_return_key=True)
                    msg = dh.shared_key.encode("utf-8")
                    hex_res1 = hmac.new(key=key, msg=msg,
                                        digestmod=digestmod).hexdigest()
                    current_noise = int(hex_res1[0:8], 16) / scale
                    if i > self._trainer_id:
                        noise = noise + current_noise
                    else:
                        noise = noise - current_noise

            scope = fluid.global_scope()
            for param_name in self._param_name_list:
                fluid.global_scope().var(param_name + str(
                    self._trainer_id)).get_tensor().set(
                        numpy.array(
                            scope.find_var(param_name + str(self._trainer_id))
                            .get_tensor()) + noise, fluid.CPUPlace())
            self.exe.run(self._send_program)
        self.cur_step += 1
        return loss



#   Copyright (c) 2019 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import sys
import math
from functools import reduce

import collections
import six
import logging

import numpy as np
import paddle.fluid as fluid

from .details.ps_dispatcher import RoundRobin, PSDispatcher
from paddle.fluid import core, framework, unique_name
from paddle.fluid.framework import Program, default_main_program, \
    default_startup_program, Block, Parameter, grad_var_name
from .details import wait_server_ready, UnionFind, VarStruct, VarsDistributed
from .details import delete_ops, find_op_by_output_arg
from paddle.fluid.distribute_lookup_table import find_distributed_lookup_table
from paddle.fluid.transpiler.distribute_transpiler import DistributeTranspilerConfig, slice_variable

LOOKUP_TABLE_TYPE = "lookup_table"
LOOKUP_TABLE_GRAD_TYPE = "lookup_table_grad"
OP_ROLE_VAR_ATTR_NAME = core.op_proto_and_checker_maker.kOpRoleVarAttrName()
RPC_OP_ROLE_ATTR_NAME = op_role_attr_name = core.op_proto_and_checker_maker.kOpRoleAttrName(
)
OPT_OP_ROLE_ATTR_VALUE = core.op_proto_and_checker_maker.OpRole.Optimize
RPC_OP_ROLE_ATTR_VALUE = core.op_proto_and_checker_maker.OpRole.RPC
DIST_OP_ROLE_ATTR_VALUE = core.op_proto_and_checker_maker.OpRole.Dist
LR_SCHED_OP_ROLE_ATTR_VALUE = core.op_proto_and_checker_maker.OpRole.LRSched
PRINT_LOG = False


def log(*args):
    if PRINT_LOG:
        print(args)


def same_or_split_var(p_name, var_name):
    return p_name == var_name or p_name.startswith(var_name + ".block")


class FLDistributeTranspiler(object):
    """
    **FlDistributeTranspiler**

    Convert the fluid program to distributed data-parallelism programs.

    In pserver mode, the trainers' main program do forward, backward and optimizaiton.
    pserver's main_program will sum and scale.


    Examples:
        .. code-block:: python

            x = fluid.layers.data(name='x', shape=[13], dtype='float32')
            y = fluid.layers.data(name='y', shape=[1], dtype='float32')
            y_predict = fluid.layers.fc(input=x, size=1, act=None)

            cost = fluid.layers.square_error_cost(input=y_predict, label=y)
            avg_loss = fluid.layers.mean(cost)

            sgd_optimizer = fluid.optimizer.SGD(learning_rate=0.001)
            sgd_optimizer.minimize(avg_loss)

            # for pserver mode
            pserver_endpoints = "192.168.0.1:6174,192.168.0.2:6174"
            trainer_endpoints = "192.168.0.1:6174,192.168.0.2:6174"
            current_endpoint = "192.168.0.1:6174"
            trainer_id = 0
            trainers = 4
            role = "PSERVER"
            t = fluid.FlDistributeTranspiler()
            t.transpile(
                 trainer_id, pservers=pserver_endpoints, trainers=trainers)
            if role == "PSERVER":
                 pserver_program = t.get_pserver_program(current_endpoint)
                 pserver_startup_program = t.get_startup_program(current_endpoint,
                                                                pserver_program)
            elif role == "TRAINER":
                 trainer_program = t.get_trainer_program()

    """

    def __init__(self, config=None):
        if config is not None:
            self.config = config
        else:
            self.config = DistributeTranspilerConfig()

        if self.config.split_method is None:
            self.config.split_method = RoundRobin

        global PRINT_LOG
        if self.config.print_log:
            PRINT_LOG = True
        assert (self.config.min_block_size >= 8192)
        assert (self.config.split_method.__bases__[0] == PSDispatcher)

    def _get_all_remote_sparse_update_op(self, main_program):
        sparse_update_ops = []
        sparse_update_op_types = [
            "lookup_table", "nce", "hierarchical_sigmoid"
        ]
        for op in main_program.global_block().ops:
            if op.type in sparse_update_op_types and op.attr(
                    'remote_prefetch') is True:
                sparse_update_ops.append(op)
        return sparse_update_ops

    def transpile(self,
                  trainer_id,
                  program=None,
                  pservers="127.0.0.1:6174",
                  trainers=1,
                  sync_mode=True,
                  startup_program=None,
                  current_endpoint="127.0.0.1:6174"):
        """
        Run the transpiler. Transpile the input program.

        Args:
            trainer_id (int): id for current trainer worker, if you have
                n workers, the id may range from 0 ~ n-1
            program (Program|None): program to transpile,
                default is fluid.default_main_program().
            startup_program (Program|None): startup_program to transpile,
                default is fluid.default_startup_program().
            pservers (str): comma separated ip:port string for the pserver
                list.
            trainers (int|str): in pserver mode this is the number of
                trainers.
            sync_mode (bool): Do sync training or not, default is True.
            startup_program (Program|None): startup_program to transpile,
                default is fluid.default_main_program().
            current_endpoint (str): In pserver mode
                this argument is not used.

        Examples:
            .. code-block:: python

                transpiler = fluid.DistributeTranspiler()
                t.transpile(
                    trainer_id=0,
                    pservers="127.0.0.1:7000,127.0.0.1:7001",
                    trainers=2,
                    sync_mode=False,
                    current_endpoint="127.0.0.1:7000")
        """
        if program is None:
            program = default_main_program()
        if startup_program is None:
            startup_program = default_startup_program()
        self.origin_program = program
        self.startup_program = startup_program
        self.origin_startup_program = self.startup_program.clone()

        self.trainer_num = trainers
        self.sync_mode = sync_mode
        self.trainer_id = trainer_id
        pserver_endpoints = pservers.split(",")
        self.pserver_endpoints = pserver_endpoints
        self.vars_overview = VarsDistributed()
        self.optimize_ops, self.params_grads = self._get_optimize_pass()

        ps_dispatcher = self.config.split_method(self.pserver_endpoints)
        self.table_name = find_distributed_lookup_table(self.origin_program)
        self.has_distributed_lookup_table = self.table_name != None
        self.param_name_to_grad_name = dict()
        self.grad_name_to_param_name = dict()
        for param_var, grad_var in self.params_grads:
            self.param_name_to_grad_name[param_var.name] = grad_var.name
            self.grad_name_to_param_name[grad_var.name] = param_var.name

        # get all sparse update ops
        self.sparse_update_ops = self._get_all_remote_sparse_update_op(
            self.origin_program)
        # use_sparse_update_param_name -> split_height_section
        self.sparse_param_to_height_sections = dict()

        # add distributed attrs to program
        self.origin_program._is_distributed = True
        self.origin_program._endpoints = self.pserver_endpoints
        self.origin_program._ps_endpoint = current_endpoint
        self.origin_program._is_chief = self.trainer_id == 0
        self.origin_program._distributed_lookup_table = self.table_name if self.table_name else None

        # split and create vars, then put splited vars in dicts for later use.
        # step 1: split and create vars, then put splited vars in dicts for later use.
        self._init_splited_vars()

        # step 2: insert send op to send gradient vars to parameter servers
        ps_dispatcher.reset()
        send_vars = []

        # in general cases, the number of pservers is times of 2, and this
        # will lead to uneven distribution among weights and bias:
        #       fc_w@GRAD_trainer_0, fc_w@GRAD_trainer_1 --> pserver1
        #       fc_b@GRAD_trainer_0, fc_b@GRAD_trainer_1 --> pserver2
        # shuffle the map will avoid the uneven distribution above

        self.opti_name_to_send_dummy_out = dict()
        self.recv_program = self.origin_program.clone()
        all_ops = []
        for op in self.recv_program.global_block().ops:
            all_ops.append(op)
        delete_ops(self.recv_program.global_block(), all_ops)

        self.split_num = len(program.global_block().ops)
        for opti_varname in self._opti_var_list:
            opti_var = program.global_block().var(opti_varname)
            eplist = ps_dispatcher.dispatch([opti_var])

            dummy_output = program.global_block().create_var(
                name=framework.generate_control_dev_var_name())
            self.opti_name_to_send_dummy_out[opti_varname] = dummy_output

            program.global_block().append_op(
                type="send",
                inputs={"X": [opti_var]},
                outputs={"Out": dummy_output},
                attrs={
                    "epmap": eplist,
                    RPC_OP_ROLE_ATTR_NAME: RPC_OP_ROLE_ATTR_VALUE,
                    OP_ROLE_VAR_ATTR_NAME:
                    [self._opti_to_param[opti_varname], opti_varname],
                    "sync_mode": not self.sync_mode,
                })
            send_vars.append(opti_var)

        if self.sync_mode:
            send_barrier_out = program.global_block().create_var(
                name=framework.generate_control_dev_var_name())
            input_deps = list(self.opti_name_to_send_dummy_out.values())

            program.global_block().append_op(
                type="send_barrier",
                inputs={"X": list(input_deps)},
                outputs={"Out": send_barrier_out},
                attrs={
                    "endpoints": pserver_endpoints,
                    "sync_mode": self.sync_mode,
                    "trainer_id": self.trainer_id,
                    RPC_OP_ROLE_ATTR_NAME: RPC_OP_ROLE_ATTR_VALUE
                })

        # step 3: insert recv op to receive parameters from parameter server
        recv_vars = []
        for _, var in enumerate(send_vars):
            recv_vars.append(program.global_block().var(self._opti_to_param[
                var.name]))
        ps_dispatcher.reset()
        eplist = ps_dispatcher.dispatch(recv_vars)
        for i, ep in enumerate(eplist):
            self.param_grad_ep_mapping[ep]["params"].append(recv_vars[i])
            self.param_grad_ep_mapping[ep]["opti"].append(send_vars[i])

            distributed_var = self.vars_overview.get_distributed_var_by_slice(
                recv_vars[i].name)
            distributed_var.endpoint = ep

        # step4: Concat the parameters splits together after recv.
        all_recv_outputs = []
        for opti_varname in self._opti_var_list:
            opti_var = program.global_block().var(opti_varname)
            param_varname = self._opti_to_param[opti_varname]
            param_var = program.global_block().var(param_varname)
            eps = []
            table_names = []
            index = [v.name for v in recv_vars].index(param_varname)
            eps.append(eplist[index])
            table_names.append(var.name)
            if self.sync_mode:
                recv_dep_in = send_barrier_out
            # get recv op_role_var, if not splited, the grad should have .trainer suffix
            # if splited, grad should be the original grad var name. ParallelExecutor
            # will use op_role_var to get expected device place to run this op.

            all_recv_outputs.extend([param_var])
            self.recv_program.global_block().append_op(
                type="recv",
                inputs={"X": []},
                outputs={"Out": [param_var]},
                attrs={
                    "epmap": eps,
                    "trainer_id": self.trainer_id,
                    RPC_OP_ROLE_ATTR_NAME: RPC_OP_ROLE_ATTR_VALUE,
                    OP_ROLE_VAR_ATTR_NAME: [param_varname, opti_varname],
                    "sync_mode": not self.sync_mode
                })

        if self.sync_mode:
            # form a WAW dependency
            self.recv_program.global_block()._insert_op(
                index=len(self._opti_var_list),
                type="fetch_barrier",
                inputs={},
                outputs={"Out": all_recv_outputs},
                attrs={
                    "endpoints": pserver_endpoints,
                    "trainer_id": self.trainer_id,
                    RPC_OP_ROLE_ATTR_NAME: RPC_OP_ROLE_ATTR_VALUE
                })

        self._get_trainer_startup_program(recv_vars=recv_vars, eplist=eplist)

        self._get_distributed_optimizer_vars()
        self.origin_program._parameters_on_pservers = self.vars_overview

    def get_trainer_program(self, wait_port=True):
        """
        Get transpiled trainer side program.

        Returns:
            Program: trainer side program.
        """
        # remove optimize ops and add a send op to main_program
        # FIXME(typhoonzero): Also ops like clip_gradient, lrn_decay?

        lr_ops = self._get_lr_ops()

        self.origin_program.__str__()

        self.send_program = self.origin_program.clone()
        compute_ops = self.send_program.global_block().ops[0:self.split_num]
        delete_ops(self.send_program.global_block(), compute_ops)
        send_ops = self.origin_program.global_block().ops[self.split_num:]
        delete_ops(self.origin_program.global_block(), send_ops)

        return self.recv_program, self.origin_program, self.send_program

    def _get_trainer_startup_program(self, recv_vars, eplist):
        """
        Get transpiled trainer side startup program.

        Args:
            recv_vars (list): Variable list to recv for current trainer_id
            eplist (list): A list of strings indicating

        Returns:
            Program: trainer side startup program.
        """
        startup_program = self.startup_program

        # FIXME(gongwb): delete not need ops.
        # note that: some parameter is not trainable and those ops can't be deleted.
        for opti_varname in self._opti_var_list:
            opti_var = self.origin_program.global_block().var(opti_varname)
            param_varname = self._opti_to_param[opti_varname]
            var = self.origin_program.global_block().var(param_varname)

            # Get the eplist of recv vars
            eps = []
            table_names = []
            index = [v.name for v in recv_vars].index(param_varname)
            eps.append(eplist[index])

        return startup_program

    def get_pserver_program(self, endpoint):
        """
        Get parameter server side program.

        Args:
            endpoint (str): current parameter server endpoint.

        Returns:
            Program: the program for current parameter server to run.
        """
        # TODO(panyx0718): Revisit this assumption. what if #blocks > #pservers.
        # NOTE: assume blocks of the same variable is not distributed
        # on the same pserver, only change param/grad varnames for
        # trainers to fetch.
        sys.stderr.write(
            "get_pserver_program() is deprecated, call get_pserver_programs() to get pserver main and startup in a single call.\n"
        )
        # step1
        pserver_program = Program()
        pserver_program.random_seed = self.origin_program.random_seed
        pserver_program._copy_dist_param_info_from(self.origin_program)

        # step2: Create vars to receive vars at parameter servers.
        recv_inputs = []
        for v in self.param_grad_ep_mapping[endpoint]["params"]:
            self._clone_var(pserver_program.global_block(), v)
        for v in self.param_grad_ep_mapping[endpoint]["opti"]:
            # create vars for each trainer in global scope, so
            # we don't need to create them when grad arrives.
            # change client side var name to origin name by
            # removing ".trainer_%d" suffix
            suff_idx = v.name.find(".opti.trainer_")
            if suff_idx >= 0:
                orig_var_name = v.name[:suff_idx]
            # NOTE: single_trainer_var must be created for multi-trainer
            # case to merge grads from multiple trainers
            single_trainer_var = pserver_program.global_block().var(
                orig_var_name)

            if self.sync_mode and self.trainer_num > 1:
                for trainer_id in range(self.trainer_num):
                    var = pserver_program.global_block().create_var(
                        name="%s.opti.trainer_%d" %
                        (orig_var_name, trainer_id),
                        persistable=False,
                        type=v.type,
                        dtype=v.dtype,
                        shape=v.shape)
                    recv_inputs.append(var)

        # step 3
        # Create a union-find data structure from optimize ops,
        # If two ops are connected, we could add these two ops
        # into one set.
        ufind = self._create_ufind(self.optimize_ops)
        # step 3.2
        # Iterate through the ops and append optimize op which
        # located on current pserver
        opt_op_on_pserver = []
        for _, op in enumerate(self.optimize_ops):
            if self._is_optimizer_op(op) and self._is_opt_op_on_pserver(
                    endpoint, op):
                opt_op_on_pserver.append(op)

        # step 3.4
        # Iterate through the ops, and if an op and the optimize ops
        # which located on current pserver are in one set, then
        # append it into the sub program.

        global_ops = []

        # sparse grad name to param name
        sparse_grad_to_param = []

        # append lr decay ops to the child block if exists
        lr_ops = self._get_lr_ops()
        # record optimize blocks and we can run them on pserver parallel
        opti_blocks = []

        # append op to the current block
        grad_to_block_id = []
        pre_block_idx = pserver_program.num_blocks - 1
        for idx, opt_op in enumerate(self._opti_var_list):
            per_opt_block = pserver_program._create_block(pre_block_idx)
            opti_blocks.append(per_opt_block)
            optimize_target_param_name = self._opti_to_param[opt_op]
            pserver_block = per_opt_block.program.global_block()
            # append grad merging ops before clip and weight decay
            # e.g. merge grad -> L2Decay op -> clip op -> optimize
            merged_var = pserver_block.vars[optimize_target_param_name]
            if self.sync_mode and self.trainer_num > 1:
                vars2merge = []
                for i in range(self.trainer_num):
                    per_trainer_name = "%s.opti.trainer_%d" % \
                                       (optimize_target_param_name, i)
                    vars2merge.append(pserver_block.vars[per_trainer_name])
                per_opt_block.append_op(
                    type="sum",
                    inputs={"X": vars2merge},
                    outputs={"Out": merged_var},
                    attrs={"use_mkldnn": False})
                per_opt_block.append_op(
                    type="scale",
                    inputs={"X": merged_var},
                    outputs={"Out": merged_var},
                    attrs={"scale": 1.0 / float(self.trainer_num)})

        # In some case, some parameter server will have no parameter to optimize
        # So we give an empty optimize block to parameter server.
        attrs = {
            "optimize_blocks": opti_blocks,
            "endpoint": endpoint,
            "Fanin": self.trainer_num,
            "sync_mode": self.sync_mode,
        }

        # step5 append the listen_and_serv op
        pserver_program.global_block().append_op(
            type="fl_listen_and_serv",
            inputs={'X': recv_inputs},
            outputs={},
            attrs=attrs)

        pserver_program._sync_with_cpp()
        # save pserver program to generate pserver side startup relatively.
        self.pserver_program = pserver_program
        return pserver_program

    def get_startup_program(self,
                            endpoint,
                            pserver_program=None,
                            startup_program=None):
        """
        **Deprecated**

        Get startup program for current parameter server.
        Modify operator input variables if there are variables that
        were split to several blocks.

        Args:
            endpoint (str): current pserver endpoint.
            pserver_program (Program): deprecated, call get_pserver_program first.
            startup_program (Program): deprecated, should pass startup_program
                when initalizing

        Returns:
            Program: parameter server side startup program.
        """
        s_prog = Program()
        orig_s_prog = self.startup_program
        s_prog.random_seed = orig_s_prog.random_seed
        params = self.param_grad_ep_mapping[endpoint]["params"]

        def _get_splited_name_and_shape(varname):
            for idx, splited_param in enumerate(params):
                pname = splited_param.name
                if same_or_split_var(pname, varname) and varname != pname:
                    return pname, splited_param.shape
            return "", []

        # 1. create vars in pserver program to startup program
        pserver_vars = pserver_program.global_block().vars
        created_var_map = collections.OrderedDict()
        for _, var in six.iteritems(pserver_vars):
            tmpvar = s_prog.global_block()._clone_variable(var)
            created_var_map[var.name] = tmpvar

        # 2. rename op outputs
        for op in orig_s_prog.global_block().ops:
            new_outputs = collections.OrderedDict()
            # do not append startup op if var is not on this pserver
            op_on_pserver = False
            # TODO(gongwb): remove this line.
            if op.type not in ["recv", "fetch_barrier", "concat"]:
                for key in op.output_names:
                    newname, _ = _get_splited_name_and_shape(op.output(key)[0])
                    if newname:
                        op_on_pserver = True
                        new_outputs[key] = created_var_map[newname]
                    elif op.output(key)[0] in pserver_vars:
                        op_on_pserver = True
                        new_outputs[key] = pserver_vars[op.output(key)[0]]

            if op_on_pserver:
                # most startup program ops have no inputs
                new_inputs = self._get_input_map_from_op(pserver_vars, op)

                if op.type in [
                        "gaussian_random", "fill_constant", "uniform_random",
                        "truncated_gaussian_random"
                ]:
                    op._set_attr("shape", list(new_outputs["Out"].shape))
                s_prog.global_block().append_op(
                    type=op.type,
                    inputs=new_inputs,
                    outputs=new_outputs,
                    attrs=op.all_attrs())

        return s_prog

    # ====================== private transpiler functions =====================
    def _get_slice_var_info(self, slice_var):
        block_suffix = "block"
        block_idx = 0
        offset = 0
        is_slice = False

        orig_var_name, block_name, _ = self._get_varname_parts(slice_var.name)

        if not block_name:
            return is_slice, block_idx, offset

    def _get_distributed_optimizer_vars(self):
        def _get_distributed_optimizer_var(endpoint):
            opt_op_on_pserver = []
            for _, op in enumerate(self.optimize_ops):
                if self._is_optimizer_op(op) and self._is_opt_op_on_pserver(
                        endpoint, op):
                    opt_op_on_pserver.append(op)

            for opt_op in opt_op_on_pserver:
                dist_var = None
                for key in opt_op.input_names:
                    if key == "Param":
                        param_name = opt_op.input(key)[0]
                        dist_var = self.vars_overview.get_distributed_var_by_origin_and_ep(
                            param_name, endpoint)
                        break
                for key in opt_op.input_names:
                    if key in ["Param", "Grad", "LearningRate"]:
                        continue

        for ep in self.pserver_endpoints:
            _get_distributed_optimizer_var(ep)

    def _update_dist_lookup_table_vars(self, param_list, grad_list,
                                       params_grads):
        # TODO(wuyi): put find a way to put dist lookup table stuff all together.
        # update self.table_param_grad and self.trainer_side_table_grad_list
        program = self.origin_program
        return param_list, grad_list

    def _init_splited_vars(self):
        # update these mappings for further transpile:
        # 1. param_var_mapping: param var name -> [splited params vars]
        # 2. grad_var_mapping: grad var name -> [splited grads vars]
        # 3. grad_param_mapping: grad.blockx -> param.blockx
        # 4. param_grad_ep_mapping: ep -> {"params": [], "grads": []}

        param_list = []
        grad_list = []
        param_grad_set = set()
        for p, g in self.params_grads:
            # skip parameter marked not trainable
            if type(p) == Parameter and p.trainable == False:
                continue
            if p.name not in param_grad_set:
                param_list.append(p)
                param_grad_set.add(p.name)
            if g.name not in param_grad_set:
                grad_list.append(g)
                param_grad_set.add(g.name)

        # To do : consider lookup table later
        param_list, grad_list = self._update_dist_lookup_table_vars(
            param_list, grad_list, self.params_grads)

        if self.config.slice_var_up:
            # when we slice var up into blocks, we will slice the var according to
            # pserver services' count. A pserver may have two or more listening ports.
            grad_blocks = slice_variable(grad_list,
                                         len(self.pserver_endpoints),
                                         self.config.min_block_size)
            param_blocks = slice_variable(param_list,
                                          len(self.pserver_endpoints),
                                          self.config.min_block_size)
        assert (len(grad_blocks) == len(param_blocks))

        # origin_param_name -> [splited_param_vars]
        self.param_var_mapping = self._create_vars_from_blocklist(
            self.origin_program, param_blocks)

        for orig_name, splited_vars in self.param_var_mapping.items():
            orig_var = self.origin_program.global_block().var(orig_name)
            for splited_var in splited_vars:
                is_slice, block_id, offset = self._get_slice_var_info(
                    splited_var)

                self.vars_overview.add_distributed_var(
                    origin_var=orig_var,
                    slice_var=splited_var,
                    block_id=block_id,
                    offset=offset,
                    is_slice=is_slice,
                    vtype="Param")

        # origin_grad_name -> [splited_grad_vars]
        self.grad_var_mapping = self._create_vars_from_blocklist(
            self.origin_program, grad_blocks)
        #add_trainer_suffix=self.trainer_num > 1)
        # dict(grad_splited_var -> param_splited_var)
        self.grad_param_mapping = collections.OrderedDict()
        for g, p in zip(grad_blocks, param_blocks):
            g_name, g_bid, _ = g.split(":")
            p_name, p_bid, _ = p.split(":")
            self.grad_param_mapping[self.grad_var_mapping[g_name][int(g_bid)]] = \
                self.param_var_mapping[p_name][int(p_bid)]

        # create mapping of endpoint -> split var to create pserver side program
        self.param_grad_ep_mapping = collections.OrderedDict()
        [
            self.param_grad_ep_mapping.update({
                ep: {
                    "params": [],
                    "opti": []
                }
            }) for ep in self.pserver_endpoints
        ]

        opti_list = []
        opti_to_param = dict()
        param_to_opti = dict()
        for op in self.optimize_ops:
            if (op.type == "sgd") or (op.type == "adam"):
                origin_name = op.output("ParamOut")
                var = self.origin_program.global_block().var(origin_name[0])
                new_var_name = "%s.opti.trainer_%d" % (origin_name[0],
                                                       self.trainer_id)
                self.origin_program.global_block().create_var(
                    name=new_var_name,
                    persistable=True,
                    shape=var.shape,
                    dtype=var.dtype,
                    type=var.type,
                    lod_level=var.lod_level)
                new_var = self.origin_program.global_block().var(new_var_name)
                opti_list.append(new_var.name)
                opti_to_param[new_var.name] = var.name
                param_to_opti[var.name] = new_var.name
                self.origin_program.global_block().append_op(
                    type="scale",
                    inputs={"X": var},
                    outputs={"Out": new_var},
                    attrs={"scale": 1.0})
        self._param_to_opti = param_to_opti
        self._opti_to_param = opti_to_param
        self._opti_var_list = opti_list

    def _create_vars_from_blocklist(self,
                                    program,
                                    block_list,
                                    add_trainer_suffix=False):
        """
        Create vars for each split.
        NOTE: only grads need to be named for different trainers, use
              add_trainer_suffix to rename the grad vars.
        Args:
            program (ProgramDesc): ProgramDesc which gradients blong.
            block_list (list[(varname, block_id, block_size)]): List of gradient blocks.
            add_trainer_suffix (Bool): Add trainer suffix to new variable's name if set True.
        Returns:
            var_mapping (collections.OrderedDict(varname->[new_varname_variable])):A dict mapping
                from original var name to each var split.
        """

        # varname->[(block_id, current_block_size)]
        block_map = collections.OrderedDict()

        var_mapping = collections.OrderedDict()
        for block_str in block_list:
            varname, offset, size = block_str.split(":")
            if varname not in block_map:
                block_map[varname] = []
            block_map[varname].append((int(offset), int(size)))

        for varname, splited in six.iteritems(block_map):
            orig_var = program.global_block().var(varname)
            if len(splited) == 1:
                var_mapping[varname] = \
                        [program.global_block().var(orig_var.name)]
                continue
        return var_mapping

    def _clone_var(self, block, var, persistable=True):
        return block.create_var(
            name=var.name,
            shape=var.shape,
            dtype=var.dtype,
            type=var.type,
            lod_level=var.lod_level,
            persistable=persistable)

    def _get_varname_parts(self, varname):
        # returns origin, blockid, trainerid
        orig_var_name = ""
        trainer_part = ""
        block_part = ""
        trainer_idx = varname.find(".trainer_")
        if trainer_idx >= 0:
            trainer_part = varname[trainer_idx + 1:]
        else:
            trainer_idx = len(varname)
        block_index = varname.find(".block")
        if block_index >= 0:
            block_part = varname[block_index + 1:trainer_idx]
        else:
            block_index = len(varname)
        orig_var_name = varname[0:min(block_index, trainer_idx)]
        return orig_var_name, block_part, trainer_part

    def _is_op_connected(self, op1, op2):
        # If one op's input is another op's output or
        # one op's output is another op's input, we say
        # the two operator is connected.
        if set(op1.desc.output_arg_names()) & set(op2.desc.input_arg_names()) or \
                set(op1.desc.input_arg_names()) & set(op2.desc.output_arg_names()):
            return True
        return False

    def _create_ufind(self, optimize_ops):
        # Create a unit find data struct by optimize ops
        ufind = UnionFind(optimize_ops)
        for i in range(len(optimize_ops)):
            for j in range(i, len(optimize_ops)):
                op1 = optimize_ops[i]
                op2 = optimize_ops[j]
                if self._is_op_connected(op1, op2):
                    ufind.union(op1, op2)
        return ufind

    def _is_optimizer_op(self, op):
        if "Param" in op.input_names and \
                "LearningRate" in op.input_names:
            return True
        return False

    def _is_opt_op_on_pserver(self, endpoint, op):
        param_names = [
            p.name for p in self.param_grad_ep_mapping[endpoint]["params"]
        ]
        if op.input("Param")[0] in param_names:
            return True

    def _get_input_map_from_op(self, varmap, op):
        """Returns a dict from op input name to the vars in varmap."""
        iomap = collections.OrderedDict()
        return iomap

    def _get_lr_ops(self):
        lr_ops = []
        block = self.origin_program.global_block()
        for op in block.ops:
            role_id = int(op.attr(RPC_OP_ROLE_ATTR_NAME))
        return lr_ops

    def _is_opt_role_op(self, op):
        # NOTE: depend on oprole to find out whether this op is for
        # optimize
        op_maker = core.op_proto_and_checker_maker
        optimize_role = core.op_proto_and_checker_maker.OpRole.Optimize
        if op_maker.kOpRoleAttrName() in op.attr_names and \
                int(op.all_attrs()[op_maker.kOpRoleAttrName()]) == int(optimize_role):
            return True
        return False

    def _get_optimize_pass(self):
        """
        Get optimizer operators, parameters and gradients from origin_program
        Returns:
            opt_ops (list): optimize operators.
            params_grads (dict): parameter->gradient.
        """
        block = self.origin_program.global_block()
        opt_ops = []
        params_grads = []
        # tmp set to dedup
        optimize_params = set()
        origin_var_dict = self.origin_program.global_block().vars
        for op in block.ops:
            if self._is_opt_role_op(op):
                opt_ops.append(op)
                if op.attr(OP_ROLE_VAR_ATTR_NAME):
                    param_name = op.attr(OP_ROLE_VAR_ATTR_NAME)[0]
                    grad_name = op.attr(OP_ROLE_VAR_ATTR_NAME)[1]
                    if not param_name in optimize_params:
                        optimize_params.add(param_name)
                        log("adding param_grad pair: ", param_name, grad_name)
                        params_grads.append([
                            origin_var_dict[param_name],
                            origin_var_dict[grad_name]
                        ])
        return opt_ops, params_grads



# coding=utf-8

#
# The MIT License (MIT)
#
# Copyright (c) 2016 Chris von Csefalvay
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
# associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
# NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
"""
diffiehellmann declares the main key exchange class.
"""

__version__ = '0.13.3'

from hashlib import sha256

from .decorators import requires_private_key
from .exceptions import MalformedPublicKey, RNGError
from .primes import PRIMES
import os

#python3
try:
    from ssl import RAND_bytes
    rng = RAND_bytes
except (AttributeError, ImportError):
    rng = os.urandom


class DiffieHellman:
    """
    Implements the Diffie-Hellman key exchange protocol.

    """

    def __init__(self, group=18, key_length=640):

        self.key_length = max(200, key_length)
        self.generator = PRIMES[group]["generator"]
        self.prime = PRIMES[group]["prime"]

    def load_private_key(self, priv_key_filepath="priv_key.txt"):
        f = open(priv_key_filepath, "r")
        self.private_key = int(f.read())
        # self.private_key = 1236621350910932696206938487330072474688096146032487063733488274339542368951034578546908207981931

    def generate_private_key(self):
        """
        Generates a private key of key_length bits and attaches it to the object as the __private_key variable.

        :return: void
        :rtype: void
        """
        key_length = self.key_length // 8 + 8
        key = 0

        try:
            key = int.from_bytes(rng(key_length), byteorder='big')
        except:
            key = int(hex(rng(key_length)), base=16)

        self.private_key = key

    def verify_public_key(self, other_public_key):
        return self.prime - 1 > other_public_key > 2 and pow(
            other_public_key, (self.prime - 1) // 2, self.prime) == 1

    @requires_private_key
    def generate_public_key(self):
        """
        Generates public key.

        :return: void
        :rtype: void
        """
        self.public_key = pow(self.generator, self.private_key, self.prime)

    @requires_private_key
    def generate_shared_secret(self, other_public_key, echo_return_key=False):
        """
        Generates shared secret from the other party's public key.

        :param other_public_key: Other party's public key
        :type other_public_key: int
        :param echo_return_key: Echo return shared key
        :type bool
        :return: void
        :rtype: void
        """
        if self.verify_public_key(other_public_key) is False:
            raise MalformedPublicKey

        self.shared_secret = pow(other_public_key, self.private_key,
                                 self.prime)
        try:
            #python3
            shared_secret_as_bytes = self.shared_secret.to_bytes(
                self.shared_secret.bit_length() // 8 + 1, byteorder='big')
        except:
            #python2
            length = self.shared_secret.bit_length() // 8 + 1
            shared_secret_as_bytes = ('%%0%dx' % (
                length << 1) % self.shared_secret).decode('hex')[-length:]
        _h = sha256()
        _h.update(bytes(shared_secret_as_bytes))

        self.shared_key = _h.hexdigest()

        if echo_return_key is True:
            return self.shared_key



# Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import print_function
from paddle.fluid.framework import Variable


class VarStruct(object):
    """
    record part properties of a Variable in python.
    """

    def __init__(self, name, shape, dtype, type, lod_level, persistable):
        self.name = name
        self.shape = shape
        self.dtype = dtype
        self.type = type
        self.lod_level = lod_level
        self.persistable = persistable


class VarDistributed(object):
    """
    a class to record the var distributed on parameter servers.
    the class will record the relationship between origin var and slice var.
    the slice var's properties, such as type/shape/offset/endpoint.
    """

    def __init__(self,
                 origin_var,
                 slice_var,
                 is_slice=None,
                 block_id=None,
                 offset=None,
                 vtype=None,
                 endpoint=None):
        """
        Args:
            origin_var(Variable|VarStruct): origin var properties
            slice_var(Variable|VarStruct): slice var properties
            is_slice(bool|None): slice or not, slice_var=True/False and its block size > 8192 are the judgement standard.
            block_id(int|None): the number about the slice var.
            offset(int|None): if the slice var is sliced, offset is the numel before the var.
            vtype(str|None): a tag, such as Optimizer/Param/RemoteProfetch.
            endpoint(str|None): which parameter the slice var on, such as "127.0.0.1:1001"
        """

        if isinstance(origin_var, Variable):
            self.origin = self.__create_var_struct(origin_var)
        else:
            self.origin = origin_var

        if isinstance(slice_var, Variable):
            self.slice = self.__create_var_struct(slice_var)
        else:
            self.slice = slice_var

        if self.equal(self.origin, self.slice):
            self.is_slice = False
            self.block_id = 0
            self.offset = 0
        else:
            self.is_slice = True
            self.block_id = 0
            self.offset = 0

        if is_slice is not None:
            self.is_slice = is_slice
        if block_id is not None:
            self.block_id = block_id
        if offset is not None:
            self.offset = offset

        self.vtype = vtype
        self.endpoint = endpoint

    @staticmethod
    def __create_var_struct(var):
        return VarStruct(var.name, var.shape, var.dtype, var.type,
                         var.lod_level, var.persistable)

    @staticmethod
    def equal(var1, var2):
        """
        the two var is equal or not.
        Returns:
            bool: equal will return True else False
        """
        assert isinstance(var1, VarStruct) and isinstance(var2, VarStruct)

        return var1.name == var2.name and \
               var1.type == var2.type and \
               var1.shape == var2.shape and \
               var1.dtype == var2.dtype and \
               var1.lod_level == var2.lod_level and \
               var1.persistable == var2.persistable

    def __str__(self):
        origin_var_str = "{name} : fluid.{type}.shape{shape}.astype({dtype})". \
            format(i="{", e="}", name=self.origin.name, type=self.origin.type,
                   shape=self.origin.shape, dtype=self.origin.dtype)

        slice_var_str = "{name} : fluid.{type}.shape{shape}.astype({dtype})" \
                        ".slice({is_slice}).block({block_id}).offset({offset})". \
            format(i="{", e="}", name=self.slice.name, type=self.slice.type,
                   shape=self.slice.shape, dtype=self.slice.dtype,
                   is_slice=self.is_slice, block_id=self.block_id, offset=self.offset)

        return "var owned: {}, origin var: ( {} ), slice var: ( {} ), endpoint: {} ".format(
            self.vtype, origin_var_str, slice_var_str, self.endpoint)


class VarsDistributed(object):
    """
    a gather about VarDistributed with many methods to find distributed vars.
    through the class, we can get overview about the distributed parameters on parameter servers.
    this class may centralized and convenient for developer to manage and get variable's distribute.
    other module can also use this to find variables such io.py.
    """

    def __init__(self):
        self.distributed_vars = []

    def add_distributed_var(self,
                            origin_var,
                            slice_var,
                            is_slice=None,
                            block_id=None,
                            offset=None,
                            vtype=None,
                            endpoint=None):
        """
        add distributed var in this.

        Args:
            origin_var(Variable|VarStruct): origin var properties
            slice_var(Variable|VarStruct): slice var properties
            is_slice(bool|None): slice or not, slice_var=True/False and its block size > 8192 are the judgement standard.
            block_id(int|None): the number about the slice var.
            offset(int|None): if the slice var is sliced, offset is the numel before the var.
            vtype(str|None): a tag, such as Optimizer/Param/RemoteProfetch.
            endpoint(str|None): which parameter the slice var on, such as "127.0.0.1:1001"
        Returns:
            None
        """
        self.distributed_vars.append(
            VarDistributed(origin_var, slice_var, is_slice, block_id, offset,
                           vtype, endpoint))

    def get_distributed_var_by_slice(self, var_name):
        """
        get distributed var by conditions.

        Args:
            var_name(str): slice var name, such as "w.traier0.block1"
        Returns:
            VarDistributed: distributed var.
        """
        for dist_var in self.distributed_vars:
            if dist_var.slice.name == var_name:
                return dist_var
        return None

    @staticmethod
    def equal(var1, var2):
        """
        the two var is equal or not.
        Returns:
            bool: equal will return True else False
        """
        return var1.name == var2.name and \
               var1.type == var2.type and \
               var1.shape == var2.shape and \
               var1.dtype == var2.dtype and \
               var1.lod_level == var2.lod_level and \
               var1.persistable == var2.persistable

    def get_distributed_var_by_origin_and_ep(self, origin_var_name, endpoint):
        """
        get distributed var by conditions.

        Args:
            origin_var_name(str):
            endpoint(str): the parameter endpoint, such as "127.0.0.1:1001"
        Returns:
            VarDistributed: distributed var.
        """
        for dist_var in self.distributed_vars:
            if dist_var.origin.name == origin_var_name and dist_var.endpoint == endpoint:
                return dist_var
        return None

    def get_distributed_vars_by_vtypes(self, vtypes, groupby=False):
        """
        get distributed vars by conditions.

        Args:
            vtype(str|None): distributed var's vtype, such as "Optimizer", "RemotePrefetch"
            groupby(bool|False): group by origin var or not.

        Returns:
            list: distributed var list.
            dict: distributed var map when groupby=True
        """
        vtype_vars = []
        for var in self.distributed_vars:
            if var.vtype in vtypes:
                vtype_vars.append(var)
        if not groupby:
            return vtype_vars

        params_map = {}
        for var in vtype_vars:
            origin_var_name = var.origin.name

            if origin_var_name in params_map.keys():
                optimizers = params_map.get(origin_var_name)
            else:
                optimizers = []
            optimizers.append(var)
            params_map[origin_var_name] = optimizers
        return params_map

    def get_distributed_vars_by_ep(self, endpoint, vtype=None):
        """
        get distributed vars by conditions.

        Args:
            endpoint(str): the parameter server endpoint, such as "127.0.0.1:2001"
            vtype(str|None): distributed var's vtype, such as "Optimizer", "RemotePrefetch"

        Returns:
            list: distributed var list.
        """
        endpoint_vars = []
        for var in self.distributed_vars:
            if var.endpoint == endpoint:
                endpoint_vars.append(var)
        if not vtype:
            return endpoint_vars

        vtype_vars = []
        for var in endpoint_vars:
            if var.vtype == vtype:
                vtype_vars.append(var)
        return vtype_vars

    def overview(self):
        """
        get the overview string about all params on all parameter servers.

        Returns:
            Str: overview string.

        """
        vars_str = []
        for var in self.distributed_vars:
            vars_str.append(str(var))
        return "\n".join(vars_str)



#   Copyright (c) 2019 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .fl_distribute_transpiler import FLDistributeTranspiler
from paddle.fluid.optimizer import SGD
import paddle.fluid as fluid


class FLStrategyFactory(object):
    """
    FLStrategyFactory is a FLStrategy builder
    Users can define strategy config to create different FLStrategy
    """

    def __init__(self):
        self._fed_avg = False
        self._dpsgd = False
        self._sec_agg = False
        self._inner_step = 1

    @property
    def fed_avg(self):
        return self._fed_avg

    @fed_avg.setter
    def fed_avg(self, s):
        self._fed_avg = s

    @property
    def dpsgd(self):
        return self._dpsgd

    @fed_avg.setter
    def dpsgd(self, s):
        self._dpsgd = s

    @property
    def sec_agg(self):
        return self._sec_agg

    @sec_agg.setter
    def sec_agg(self, s):
        self._sec_agg = s

    @property
    def inner_step(self):
        return self._inner_step

    @inner_step.setter
    def inner_step(self, s):
        self._inner_step = s

    def create_fl_strategy(self):
        """
        The factory method
        """
        if self._fed_avg == True:
            strategy = FedAvgStrategy()
            strategy._fed_avg = True
            strategy._dpsgd = False
            strategy._sec_agg = False
        elif self._dpsgd == True:
            strategy = DPSGDStrategy()
            strategy._fed_avg = False
            strategy._dpsgd = True
            strategy._sec_agg = False
        elif self._sec_agg == True:
            strategy = SecAggStrategy()
            strategy._fed_avg = False
            strategy._dpsgd = False
            strategy._sec_agg = True
        strategy._inner_step = self._inner_step
        return strategy


class FLStrategyBase(object):
    """
    FLStrategyBase is federated learning algorithm container
    """

    def __init__(self):
        self._fed_avg = False
        self._dpsgd = False
        self._inner_step = 1
        pass

    def minimize(self, optimizer=None, losses=[]):
        """
        minmize can do minimization as paddle.fluid.Optimizer.minimize does
        this function can be overloaded so that for some FLStrategy, the
        program should be transpiled before minimize

        Args:
            optimizer(paddle.fluid.optimizer): the user defined optimizer
            losses(List(Variable)): list of loss variables in paddle.fluid
        """
        for loss in losses:
            optimizer.minimize(loss)

    def _build_trainer_program_for_job(self,
                                       trainer_id=0,
                                       program=None,
                                       ps_endpoints=[],
                                       trainers=0,
                                       sync_mode=True,
                                       startup_program=None,
                                       job=None):
        pass

    def _build_server_programs_for_job(self,
                                       program=None,
                                       ps_endpoints=[],
                                       trainers=0,
                                       sync_mode=True,
                                       startup_program=None,
                                       job=None):
        pass


class DPSGDStrategy(FLStrategyBase):
    """
    DPSGDStrategy: Deep Learning with Differential Privacy. 2016
    """

    def __init__(self):
        super(DPSGDStrategy, self).__init__()

    @property
    def learning_rate(self):
        return self._learning_rate

    @learning_rate.setter
    def learning_rate(self, s):
        self._learning_rate = s

    @property
    def clip(self):
        return self._clip

    @clip.setter
    def clip(self, s):
        self._clip = s

    @property
    def batch_size(self):
        return self._batch_size

    @batch_size.setter
    def batch_size(self, s):
        self._batch_size = s

    @property
    def sigma(self):
        return self._sigma

    @sigma.setter
    def sigma(self, s):
        self._sigma = s

    def minimize(self, optimizer=None, losses=[]):
        """
        Define Dpsgd optimizer
        """
        optimizer = fluid.optimizer.Dpsgd(
            self._learning_rate,
            clip=self._clip,
            batch_size=self._batch_size,
            sigma=self._sigma)
        optimizer.minimize(losses[0])

    def _build_trainer_program_for_job(self,
                                       trainer_id=0,
                                       program=None,
                                       ps_endpoints=[],
                                       trainers=0,
                                       sync_mode=True,
                                       startup_program=None,
                                       job=None):
        transpiler = fluid.DistributeTranspiler()
        transpiler.transpile(
            trainer_id,
            program=program,
            pservers=",".join(ps_endpoints),
            trainers=trainers,
            sync_mode=sync_mode,
            startup_program=startup_program)
        main = transpiler.get_trainer_program(wait_port=False)
        job._trainer_startup_programs.append(startup_program)
        job._trainer_main_programs.append(main)

    def _build_server_programs_for_job(self,
                                       program=None,
                                       ps_endpoints=[],
                                       trainers=0,
                                       sync_mode=True,
                                       startup_program=None,
                                       job=None):
        transpiler = fluid.DistributeTranspiler()
        trainer_id = 0
        transpiler.transpile(
            trainer_id,
            program=program,
            pservers=",".join(ps_endpoints),
            trainers=trainers,
            sync_mode=sync_mode,
            startup_program=startup_program)
        job.set_server_endpoints(ps_endpoints)
        for endpoint in ps_endpoints:
            main_prog = transpiler.get_pserver_program(endpoint)
            startup_prog = transpiler.get_startup_program(endpoint, main_prog)
            job._server_startup_programs.append(startup_prog)
            job._server_main_programs.append(main_prog)


class FedAvgStrategy(FLStrategyBase):
    """
    FedAvgStrategy: this is model averaging optimization proposed in
    H. Brendan McMahan, Eider Moore, Daniel Ramage, Blaise Aguera y Arcas. Federated Learning of Deep Networks using Model Averaging. 2017
    """

    def __init__(self):
        super(FedAvgStrategy, self).__init__()

    def minimize(self, optimizer=None, losses=[]):
        """
        minimize the first loss as in paddle.fluid
        """
        optimizer.minimize(losses[0])

    def _build_trainer_program_for_job(self,
                                       trainer_id=0,
                                       program=None,
                                       ps_endpoints=[],
                                       trainers=0,
                                       sync_mode=True,
                                       startup_program=None,
                                       job=None):
        transpiler = FLDistributeTranspiler()
        transpiler.transpile(
            trainer_id,
            program=program,
            pservers=",".join(ps_endpoints),
            trainers=trainers,
            sync_mode=sync_mode,
            startup_program=startup_program)
        recv, main, send = transpiler.get_trainer_program()
        job._trainer_startup_programs.append(startup_program)
        job._trainer_main_programs.append(main)
        job._trainer_send_programs.append(send)
        job._trainer_recv_programs.append(recv)

    def _build_server_programs_for_job(self,
                                       program=None,
                                       ps_endpoints=[],
                                       trainers=0,
                                       sync_mode=True,
                                       startup_program=None,
                                       job=None):
        transpiler = FLDistributeTranspiler()
        trainer_id = 0
        transpiler.transpile(
            trainer_id,
            program=program,
            pservers=",".join(ps_endpoints),
            trainers=trainers,
            sync_mode=sync_mode,
            startup_program=startup_program)
        job.set_server_endpoints(ps_endpoints)
        for endpoint in ps_endpoints:
            main_prog = transpiler.get_pserver_program(endpoint)
            startup_prog = transpiler.get_startup_program(endpoint, main_prog)
            job._server_startup_programs.append(startup_prog)
            job._server_main_programs.append(main_prog)


class SecAggStrategy(FedAvgStrategy):
    """
    DPSGDStrategy: this is model averaging optimization proposed in
    Aaron Segal, Antonio Marcedone, Benjamin Kreuter, et al.
    Practical Secure Aggregation  for Privacy-Preserving Machine Learning,
    The 24th ACM Conference on Computer and Communications Security ( CCS2017 ).
    """

    def __init__(self):
        super(SecAggStrategy, self).__init__()
        self._param_name_list = []

    @property
    def param_name_list(self):
        return self._param_name_list

    @param_name_list.setter
    def param_name_list(self, s):
        self._param_name_list = s



#   Copyright (c) 2019  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Setup for pip package."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import platform

from setuptools import find_packages
from setuptools import setup
from paddle_fl.version import fl_version


def python_version():
    return [int(v) for v in platform.python_version().split(".")]


max_version, mid_version, min_version = python_version()

REQUIRED_PACKAGES = [
    'six >= 1.10.0', 'protobuf >= 3.1.0', 'paddlepaddle >= 1.6', 'zmq',
    'paddlepaddle-gpu >= 1.6'
]

if max_version < 3:
    REQUIRED_PACKAGES += ["enum"]
else:
    REQUIRED_PACKAGES += ["numpy"]

REQUIRED_PACKAGES += ["unittest2"]

setup(
    name='paddle_fl',
    version=fl_version.replace('-', ''),
    description=('Federated Deep Learning Package Based on PaddlePaddle.'),
    long_description='',
    url='https://github.com/PaddlePaddle/PaddleFL',
    author='PaddlePaddle Author',
    author_email='paddle-dev@baidu.com',
    install_requires=REQUIRED_PACKAGES,
    packages=find_packages(),
    # PyPI package information.
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    license='Apache 2.0',
    keywords=(
        'paddle_fl paddlepaddle multi-task transfer distributed-training'))



# Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import time
import socket
from contextlib import closing
from six import string_types


def wait_server_ready(endpoints):
    """
    Wait until parameter servers are ready, use connext_ex to detect
    port readiness.

    Args:
        endpoints (list): endpoints string list, like:
                         ["127.0.0.1:8080", "127.0.0.1:8081"]

    Examples:
        .. code-block:: python

           wait_server_ready(["127.0.0.1:8080", "127.0.0.1:8081"])
    """
    assert not isinstance(endpoints, string_types)
    while True:
        all_ok = True
        not_ready_endpoints = []
        for ep in endpoints:
            ip_port = ep.split(":")
            with closing(socket.socket(socket.AF_INET,
                                       socket.SOCK_STREAM)) as sock:
                sock.settimeout(2)
                result = sock.connect_ex((ip_port[0], int(ip_port[1])))
                if result != 0:
                    all_ok = False
                    not_ready_endpoints.append(ep)
        if not all_ok:
            sys.stderr.write("server not ready, wait 3 sec to retry...\n")
            sys.stderr.write("not ready endpoints:" + str(not_ready_endpoints)
                             + "\n")
            sys.stderr.flush()
            time.sleep(3)
        else:
            break
